@@ -41,7 +41,8 @@ class Connector(threading.Thread):
                  u_key, auth_key, doc_manager=None, auth_username=None,
                  collection_dump=True, batch_size=constants.DEFAULT_BATCH_SIZE,
                  fields=None, dest_mapping={},
-                 auto_commit_interval=constants.DEFAULT_COMMIT_INTERVAL):
+                 auto_commit_interval=constants.DEFAULT_COMMIT_INTERVAL,
+                 continue_on_error=False, tz_aware=False):
 
         if target_url and not doc_manager:
             raise errors.ConnectorError("Cannot create a Connector with a "
@@ -56,8 +57,8 @@ class Connector(threading.Thread):
         def load_doc_manager(path):
             name, _ = os.path.splitext(os.path.basename(path))
             try:
-                import importlib.machinery
-                loader = importlib.machinery.SourceFileLoader(name, path)
+                from importlib.machinery import SourceFileLoader
+                loader = SourceFileLoader(name, path)
                 module = loader.load_module(name)
             except ImportError:
                 module = imp.load_source(name, path)
@@ -100,6 +101,9 @@ class Connector(threading.Thread):
         #The dict of source namespace to destination namespace
         self.dest_mapping = dest_mapping
 
+        #Whether the collection dump gracefully handles exceptions
+        self.continue_on_error = continue_on_error
+
         #The key that is a unique document identifier for the target system.
         #Not necessarily the mongo unique key.
         self.u_key = u_key
@@ -126,6 +130,9 @@ class Connector(threading.Thread):
         # List of fields to export
         self.fields = fields
 
+        # Timezone awareness
+        self.tz_aware = tz_aware
+
         try:
             docman_kwargs = {"unique_key": u_key,
                              "namespace_set": ns_set,
@@ -176,13 +183,13 @@ class Connector(threading.Thread):
                     logging.critical("MongoConnector: Could not "
                                      "create a progress log: %s" %
                                      str(e))
-                    sys.exit(1)
+                    sys.exit(2)
             else:
                 if (not os.access(self.oplog_checkpoint, os.W_OK)
                         and not os.access(self.oplog_checkpoint, os.R_OK)):
                     logging.critical("Invalid permissions on %s! Exiting" %
                                      (self.oplog_checkpoint))
-                    sys.exit(1)
+                    sys.exit(2)
 
     def join(self):
         """ Joins thread, stops it from running
@@ -206,20 +213,22 @@ class Connector(threading.Thread):
         # for each of the threads write to file
         with open(self.oplog_checkpoint, 'w') as dest:
             with self.oplog_progress as oplog_prog:
-
                 oplog_dict = oplog_prog.get_dict()
-                for oplog, time_stamp in oplog_dict.items():
-                    oplog_str = str(oplog)
-                    timestamp = util.bson_ts_to_long(time_stamp)
-                    json_str = json.dumps([oplog_str, timestamp])
-                    try:
-                        dest.write(json_str)
-                    except IOError:
-                        # Basically wipe the file, copy from backup
-                        dest.truncate()
-                        with open(backup_file, 'r') as backup:
-                            shutil.copyfile(backup, dest)
-                        break
+                items = [[name, util.bson_ts_to_long(oplog_dict[name])]
+                         for name in oplog_dict]
+                if len(items) == 1:
+                    # Write 1-dimensional array, as in previous versions.
+                    json_str = json.dumps(items[0])
+                else:
+                    # Write a 2d array to support sharded clusters.
+                    json_str = json.dumps(items)
+                try:
+                    dest.write(json_str)
+                except IOError:
+                    # Basically wipe the file, copy from backup
+                    dest.truncate()
+                    with open(backup_file, 'r') as backup:
+                        shutil.copyfile(backup, dest)
 
         os.remove(self.oplog_checkpoint + '.backup')
 
@@ -239,30 +248,37 @@ class Connector(threading.Thread):
         except OSError:
             return None
 
-        source = open(self.oplog_checkpoint, 'r')
-        try:
-            data = json.load(source)
-        except ValueError:       # empty file
-            reason = "It may be empty or corrupt."
-            logging.info("MongoConnector: Can't read oplog progress file. %s" %
-                         (reason))
-            source.close()
-            return None
-
-        source.close()
-
-        count = 0
-        oplog_dict = self.oplog_progress.get_dict()
-        for count in range(0, len(data), 2):
-            oplog_str = data[count]
-            time_stamp = data[count + 1]
-            oplog_dict[oplog_str] = util.long_to_bson_ts(time_stamp)
-            #stored as bson_ts
+        with open(self.oplog_checkpoint, 'r') as progress_file:
+            try:
+                data = json.load(progress_file)
+            except ValueError:
+                logging.exception(
+                    'Cannot read oplog progress file "%s". '
+                    'It may be corrupt after Mongo Connector was shut down'
+                    'uncleanly. You can try to recover from a backup file '
+                    '(may be called "%s.backup") or create a new progress file '
+                    'starting at the current moment in time by running '
+                    'mongo-connector --no-dump <other options>. '
+                    'You may also be trying to read an oplog progress file '
+                    'created with the old format for sharded clusters. '
+                    'See https://github.com/10gen-labs/mongo-connector/wiki'
+                    '/Oplog-Progress-File for complete documentation.'
+                    % (self.oplog_checkpoint, self.oplog_checkpoint))
+                return
+            # data format:
+            # [name, timestamp] = replica set
+            # [[name, timestamp], [name, timestamp], ...] = sharded cluster
+            if not isinstance(data[0], list):
+                data = [data]
+            with self.oplog_progress:
+                self.oplog_progress.dict = dict(
+                    (name, util.long_to_bson_ts(timestamp))
+                    for name, timestamp in data)
 
     def run(self):
         """Discovers the mongo cluster and creates a thread for each primary.
         """
-        main_conn = MongoClient(self.address)
+        main_conn = MongoClient(self.address, tz_aware=self.tz_aware)
         if self.auth_key is not None:
             main_conn['admin'].authenticate(self.auth_username, self.auth_key)
         self.read_oplog_progress()
@@ -286,19 +302,17 @@ class Connector(threading.Thread):
             # Establish a connection to the replica set as a whole
             main_conn.disconnect()
             main_conn = MongoClient(self.address,
-                                    replicaSet=is_master['setName'])
+                                    replicaSet=is_master['setName'],
+                                    tz_aware=self.tz_aware)
             if self.auth_key is not None:
                 main_conn.admin.authenticate(self.auth_username, self.auth_key)
 
             #non sharded configuration
             oplog_coll = main_conn['local']['oplog.rs']
 
-            prim_admin = main_conn.admin
-            repl_set = prim_admin.command("replSetGetStatus")['set']
-
             oplog = OplogThread(
                 primary_conn=main_conn,
-                main_address=(main_conn.host + ":" + str(main_conn.port)),
+                main_address=self.address,
                 oplog_coll=oplog_coll,
                 is_sharded=False,
                 doc_manager=self.doc_managers,
@@ -306,11 +320,12 @@ class Connector(threading.Thread):
                 namespace_set=self.ns_set,
                 auth_key=self.auth_key,
                 auth_username=self.auth_username,
-                repl_set=repl_set,
+                repl_set=is_master['setName'],
                 collection_dump=self.collection_dump,
                 batch_size=self.batch_size,
                 fields=self.fields,
-                dest_mapping=self.dest_mapping
+                dest_mapping=self.dest_mapping,
+                continue_on_error=self.continue_on_error
             )
             self.shard_set[0] = oplog
             logging.info('MongoConnector: Starting connection thread %s' %
@@ -359,7 +374,8 @@ class Connector(threading.Thread):
                             dm.stop()
                         return
 
-                    shard_conn = MongoClient(hosts, replicaSet=repl_set)
+                    shard_conn = MongoClient(hosts, replicaSet=repl_set,
+                                             tz_aware=self.tz_aware)
                     oplog_coll = shard_conn['local']['oplog.rs']
 
                     oplog = OplogThread(
@@ -375,7 +391,8 @@ class Connector(threading.Thread):
                         collection_dump=self.collection_dump,
                         batch_size=self.batch_size,
                         fields=self.fields,
-                        dest_mapping=self.dest_mapping
+                        dest_mapping=self.dest_mapping,
+                        continue_on_error=self.continue_on_error
                     )
                     self.shard_set[shard_id] = oplog
                     msg = "Starting connection thread"
@@ -481,10 +498,11 @@ def main():
     #for the target system,
     parser.add_option("-u", "--unique-key", action="store", type="string",
                       dest="u_key", default="_id", help=
-                      """Used to specify the mongoDB field that will serve """
+                      """The name of the MongoDB field that will serve """
                       """as the unique key for the target system. """
-                      """The default is "_id", which can be noted by """
-                      """'-u _id'""")
+                      """Note that this option does not apply """
+                      """when targeting another MongoDB cluster. """
+                      """Defaults to "_id".""")
 
     #-f is to specify the authentication key file. This file is used by mongos
     #to authenticate connections to the shards, and we'll use it in the oplog
@@ -585,6 +603,19 @@ def main():
                       """ interval, which should be preferred to this"""
                       """ option.""")
 
+    #--continue-on-error to continue to upsert documents during a collection
+    #dump, even if the documents cannot be inserted for some reason
+    parser.add_option("--continue-on-error", action="store_true",
+                      dest="continue_on_error", default=False, help=
+                      "By default, if any document fails to upsert"
+                      " during a collection dump, the entire operation fails."
+                      " When this flag is enabled, normally fatal errors"
+                      " will be caught and logged, allowing the collection"
+                      " dump to continue.\n"
+                      "Note: Applying oplog operations to an incomplete"
+                      " set of documents due to errors may cause undefined"
+                      " behavior. Use this flag to dump only.")
+
     #-v enables vebose logging
     parser.add_option("-v", "--verbose", action="store_true",
                       dest="verbose", default=False,
@@ -595,6 +626,11 @@ def main():
                       help=("Log all output to a file rather than stream to "
                             "stderr.   Omit to stream to stderr."))
 
+    # --tz-aware enables timezone-aware datetime objects.
+    parser.add_option("--tz-aware", dest="tz_aware", action="store_true",
+                      default=False,
+                      help="Make all dates and times timezone-aware.")
+
     (options, args) = parser.parse_args()
 
     logger = logging.getLogger()
@@ -606,7 +642,7 @@ def main():
     if options.enable_syslog and options.logfile:
         print ("You cannot specify syslog and a logfile simultaneously, please"
                " choose the logging method you would prefer.")
-        sys.exit(0)
+        sys.exit(1)
 
     if options.enable_syslog:
         syslog_info = options.syslog_host.split(":")
@@ -617,11 +653,7 @@ def main():
         syslog_host.setLevel(loglevel)
         logger.addHandler(syslog_host)
     elif options.logfile is not None:
-        try:
-            log_out = logging.FileHandler(options.logfile)
-        except Exception as e:
-            raise e
-            sys.exit(0)
+        log_out = logging.FileHandler(options.logfile)
         log_out.setLevel(loglevel)
         log_out.setFormatter(logging.Formatter(
             '%(asctime)s - %(levelname)s - %(message)s'))
@@ -703,7 +735,9 @@ def main():
         batch_size=options.batch_size,
         fields=fields,
         dest_mapping=dest_mapping,
-        auto_commit_interval=options.commit_interval
+        auto_commit_interval=options.commit_interval,
+        continue_on_error=options.continue_on_error,
+        tz_aware=options.tz_aware
     )
     connector.start()
 
