@@ -40,7 +40,7 @@ class OplogThread(threading.Thread):
                  doc_manager, oplog_progress_dict, namespace_set, auth_key,
                  auth_username, repl_set=None, collection_dump=True,
                  batch_size=DEFAULT_BATCH_SIZE, fields=None,
-                 dest_mapping={}):
+                 dest_mapping={}, continue_on_error=False):
         """Initialize the oplog thread.
         """
         super(OplogThread, self).__init__()
@@ -88,22 +88,27 @@ class OplogThread(threading.Thread):
         #The dict of source namespaces to destination namespaces
         self.dest_mapping = dest_mapping
 
+        #Whether the collection dump gracefully handles exceptions
+        self.continue_on_error = continue_on_error
+
         #If authentication is used, this is an admin password.
         self.auth_key = auth_key
 
         #This is the username used for authentication.
         self.auth_username = auth_username
 
-        # List of fields to export
+        # Set of fields to export
         self.fields = fields
 
         logging.info('OplogThread: Initializing oplog thread')
 
         if is_sharded:
-            self.main_connection = MongoClient(main_address)
+            self.main_connection = MongoClient(
+                main_address, tz_aware=primary_conn.tz_aware)
         else:
-            self.main_connection = MongoClient(main_address,
-                                               replicaSet=repl_set)
+            self.main_connection = MongoClient(
+                main_address, replicaSet=repl_set,
+                tz_aware=primary_conn.tz_aware)
             self.oplog = self.main_connection['local']['oplog.rs']
 
         if auth_key is not None:
@@ -116,14 +121,26 @@ class OplogThread(threading.Thread):
             err_msg = 'OplogThread: No oplog for thread:'
             logging.warning('%s %s' % (err_msg, self.primary_connection))
 
+    @property
+    def fields(self):
+        return self._fields
+
+    @fields.setter
+    def fields(self, value):
+        if value:
+            self._fields = set(value)
+            # Always include _id field
+            self._fields.add('_id')
+        else:
+            self._fields = None
+
     def run(self):
         """Start the oplog worker.
         """
         logging.debug("OplogThread: Run thread started")
         while self.running is True:
             logging.debug("OplogThread: Getting cursor")
-            cursor = self.init_cursor()
-            logging.debug("OplogThread: Got the cursor, go go go!")
+            cursor, cursor_len = self.init_cursor()
 
             # we've fallen too far behind
             if cursor is None and self.checkpoint is not None:
@@ -133,17 +150,20 @@ class OplogThread(threading.Thread):
                 self.running = False
                 continue
 
-            #The only entry is the last one we processed
-            if cursor is None or util.retry_until_ok(cursor.count) == 1:
+            if cursor_len == 0:
                 logging.debug("OplogThread: Last entry is the one we "
                               "already processed.  Up to date.  Sleeping.")
                 time.sleep(1)
                 continue
 
+            logging.debug("OplogThread: Got the cursor, count is %d"
+                          % cursor_len)
+
             last_ts = None
             err = False
             remove_inc = 0
             upsert_inc = 0
+            update_inc = 0
             try:
                 logging.debug("OplogThread: about to process new oplog "
                               "entries")
@@ -163,51 +183,73 @@ class OplogThread(threading.Thread):
                         if entry.get("fromMigrate"):
                             continue
 
+                        # Take fields out of the oplog entry that
+                        # shouldn't be replicated. This may nullify
+                        # the document if there's nothing to do.
+                        if not self.filter_oplog_entry(entry):
+                            continue
+
                         #sync the current oplog operation
                         operation = entry['op']
                         ns = entry['ns']
 
+                        if '.' not in ns:
+                            continue
+
+                        coll = ns.split('.', 1)[1]
+                        if coll.startswith("system."):
+                            continue
+
                         # use namespace mapping if one exists
                         ns = self.dest_mapping.get(entry['ns'], ns)
 
-                        #delete
-                        try:
-                            logging.debug("OplogThread: Operation for this "
-                                          "entry is %s" % str(operation))
-                            if operation == 'd':
-                                entry['_id'] = entry['o']['_id']
-                                for dm in self.doc_managers:
+                        for docman in self.doc_managers:
+                            try:
+                                logging.debug("OplogThread: Operation for this "
+                                              "entry is %s" % str(operation))
+
+                                # Remove
+                                if operation == 'd':
+                                    entry['_id'] = entry['o']['_id']
+                                    entry['ns'] = ns
+                                    entry['_ts'] = util.bson_ts_to_long(
+                                        entry['ts'])
+                                    docman.remove(entry)
                                     remove_inc += 1
-                                    dm.remove(entry)
-                            #insert/update. They are equal because of lack
-                            #of support for partial update
-                            elif operation == 'i' or operation == 'u':
-                                doc = self.retrieve_doc(entry)
-                                if doc is not None:
+                                # Insert
+                                elif operation == 'i':  # Insert
+                                    # Retrieve inserted document from
+                                    # 'o' field in oplog record
+                                    doc = entry.get('o')
+                                    # Extract timestamp and namespace
                                     doc['_ts'] = util.bson_ts_to_long(
                                         entry['ts'])
                                     doc['ns'] = ns
-                                    for dm in self.doc_managers:
-                                        upsert_inc += 1
-                                        dm.upsert(doc)
-                        except errors.OperationFailed:
-                            logging.error(
-                                "Unable to %s doc with id %s" % (
-                                    "delete" if operation == "d" else "upsert",
-                                    str(entry['_id'])
-                                ))
-                        except errors.ConnectionFailed:
-                            logging.error(
-                                "Network error while trying to %s %s" % (
-                                    "delete" if operation == "d" else "upsert",
-                                    str(entry['_id'])
-                                ))
+                                    docman.upsert(doc)
+                                    upsert_inc += 1
+                                # Update
+                                elif operation == 'u':
+                                    doc = {"_id": entry['o2']['_id'],
+                                           "_ts": util.bson_ts_to_long(
+                                               entry['ts']),
+                                           "ns": ns}
+                                    # 'o' field contains the update spec
+                                    docman.update(doc, entry.get('o', {}))
+                                    update_inc += 1
+                            except errors.OperationFailed:
+                                logging.exception(
+                                    "Unable to process oplog document %r"
+                                    % entry)
+                            except errors.ConnectionFailed:
+                                logging.exception(
+                                    "Connection failed while processing oplog "
+                                    "document %r" % entry)
 
-                        if (remove_inc + upsert_inc) % 1000 == 0:
-                            logging.debug("OplogThread: Removed %d "
-                                          " documents and upserted %d "
-                                          " documents so far"
-                                          % (remove_inc, upsert_inc))
+                        if (remove_inc + upsert_inc + update_inc) % 1000 == 0:
+                            logging.debug(
+                                "OplogThread: Documents removed: %d, "
+                                "inserted: %d, updated: %d so far" % (
+                                    remove_inc, upsert_inc, update_inc))
 
                         logging.debug("OplogThread: Doc is processed.")
 
@@ -229,6 +271,9 @@ class OplogThread(threading.Thread):
             except (pymongo.errors.AutoReconnect,
                     pymongo.errors.OperationFailure,
                     pymongo.errors.ConfigurationError):
+                logging.exception(
+                    "Cursor closed due to an exception. "
+                    "Will attempt to reconnect.")
                 err = True
 
             if err is True and self.auth_key is not None:
@@ -247,9 +292,9 @@ class OplogThread(threading.Thread):
                 self.checkpoint = last_ts
                 self.update_checkpoint()
 
-            logging.debug("OplogThread: Sleeping.  This batch I removed %d "
-                          " documents and I upserted %d documents."
-                          % (remove_inc, upsert_inc))
+            logging.debug("OplogThread: Sleeping. Documents removed: %d, "
+                          "upserted: %d, updated: %d"
+                          % (remove_inc, upsert_inc, update_inc))
             time.sleep(2)
 
     def join(self):
@@ -259,100 +304,53 @@ class OplogThread(threading.Thread):
         self.running = False
         threading.Thread.join(self)
 
-    def retrieve_doc(self, entry):
-        """Given the doc ID's, retrieve those documents from the mongos.
+    def filter_oplog_entry(self, entry):
+        """Remove fields from an oplog entry that should not be replicated."""
+        if not self._fields:
+            return entry
+
+        def pop_excluded_fields(doc):
+            for key in set(doc) - self._fields:
+                doc.pop(key)
+
+        # 'i' indicates an insert. 'o' field is the doc to be inserted.
+        if entry['op'] == 'i':
+            pop_excluded_fields(entry['o'])
+        # 'u' indicates an update. 'o' field is the update spec.
+        elif entry['op'] == 'u':
+            pop_excluded_fields(entry['o'].get("$set", {}))
+            pop_excluded_fields(entry['o'].get("$unset", {}))
+            # not allowed to have empty $set/$unset, so remove if empty
+            if "$set" in entry['o'] and not entry['o']['$set']:
+                entry['o'].pop("$set")
+            if "$unset" in entry['o'] and not entry['o']['$unset']:
+                entry['o'].pop("$unset")
+            if not entry['o']:
+                return None
+
+        return entry
+
+    def get_oplog_cursor(self, timestamp=None):
+        """Get a cursor to the oplog after the given timestamp, filtering
+        entries not in the namespace set.
+        If no timestamp is specified, returns a cursor to the entire oplog.
         """
-
-        if not entry:
-            return None
-
-        namespace = entry['ns']
-
-        # Update operations don't have an 'o' field specifying the document
-        #- instead it specifies
-        # the changes. So we use 'o2' for updates to get the doc_id later.
-
-        if 'o2' in entry:
-            doc_field = 'o2'
-        else:
-            doc_field = 'o'
-
-        doc_id = entry[doc_field]['_id']
-        db_name, coll_name = namespace.split('.', 1)
-
-        coll = self.main_connection[db_name][coll_name]
-
-        doc = util.retry_until_ok(
-            coll.find_one,
-            {'_id': doc_id},
-            fields=self.fields
-        )
-
-        return doc
-
-    def get_oplog_cursor(self, timestamp):
-        """Move cursor to the proper place in the oplog.
-        """
-
-        logging.debug("OplogThread: Getting the oplog cursor and moving it "
-                      "to the proper place in the oplog.")
+        query = {}
+        if self.namespace_set:
+            query['ns'] = {'$in': self.namespace_set}
 
         if timestamp is None:
-            return None
+            cursor = self.oplog.find(
+                query,
+                tailable=True, await_data=True, timeout=False)
+        else:
+            query['ts'] = {'$gte': timestamp}
+            cursor = self.oplog.find(
+                query, tailable=True, await_data=True, timeout=False)
+            # Applying 8 as the mask to the cursor enables OplogReplay
+            cursor.add_option(8)
 
-        cursor, cursor_len = None, 0
-        while (True):
-            try:
-                logging.debug("OplogThread: Getting the oplog cursor "
-                              "in the while true loop for get_oplog_cursor")
-                if not self.namespace_set:
-                    cursor = self.oplog.find(
-                        {'ts': {'$gte': timestamp}},
-                        tailable=True, await_data=True
-                    )
-                else:
-                    cursor = self.oplog.find(
-                        {'ts': {'$gte': timestamp},
-                         'ns': {'$in': self.namespace_set}},
-                        tailable=True, await_data=True
-                    )
-                # Applying 8 as the mask to the cursor enables OplogReplay
-                cursor.add_option(8)
-                logging.debug("OplogThread: Cursor created, getting a count.")
-                cursor_len = cursor.count()
-                logging.debug("OplogThread: Count is %d" % cursor_len)
-                break
-            except (pymongo.errors.AutoReconnect,
-                    pymongo.errors.OperationFailure,
-                    pymongo.errors.ConfigurationError):
-                pass
-        if cursor_len == 0:
-            logging.debug("OplogThread: Initiating rollback from "
-                          "get_oplog_cursor")
-            #rollback, we are past the last element in the oplog
-            timestamp = self.rollback()
-
-            logging.info('Finished rollback')
-            return self.get_oplog_cursor(timestamp)
-        first_oplog_entry = retry_until_ok(lambda: cursor[0])
-        cursor_ts_long = util.bson_ts_to_long(first_oplog_entry.get("ts"))
-        given_ts_long = util.bson_ts_to_long(timestamp)
-        if cursor_ts_long > given_ts_long:
-            # first entry in oplog is beyond timestamp, we've fallen behind!
-            return None
-        elif cursor_len == 1:     # means we are the end of the oplog
-            self.checkpoint = timestamp
-            #to commit new TS after rollbacks
-
-            return cursor
-        elif cursor_len > 1:
-            doc = retry_until_ok(next, cursor)
-            if timestamp == doc['ts']:
-                return cursor
-            else:               # error condition
-                logging.error('OplogThread: %s Bad timestamp in config file'
-                              % self.oplog)
-                return None
+        return cursor
 
     def dump_collection(self):
         """Dumps collection into the target system.
@@ -375,7 +373,7 @@ class OplogThread(threading.Thread):
                 for coll in coll_list:
                     if coll.startswith("system"):
                         continue
-                    namespace = str(database) + "." + str(coll)
+                    namespace = "%s.%s" % (database, coll)
                     dump_set.append(namespace)
 
         timestamp = util.retry_until_ok(self.get_last_oplog_timestamp)
@@ -394,18 +392,21 @@ class OplogThread(threading.Thread):
                 # Loop to handle possible AutoReconnect
                 while attempts < 60:
                     target_coll = self.main_connection[database][coll]
+                    find_kwargs = {
+                        'fields': self._fields,
+                        'sort': [('_id', pymongo.ASCENDING)],
+                        'timeout': False
+                    }
+                    if not self.is_sharded:
+                        find_kwargs['exhaust'] = True
                     if not last_id:
                         cursor = util.retry_until_ok(
-                            target_coll.find,
-                            fields=self.fields,
-                            sort=[("_id", pymongo.ASCENDING)]
-                        )
+                            target_coll.find, **find_kwargs)
                     else:
                         cursor = util.retry_until_ok(
                             target_coll.find,
                             {"_id": {"$gt": last_id}},
-                            fields=self.fields,
-                            sort=[("_id", pymongo.ASCENDING)]
+                            **find_kwargs
                         )
                     try:
                         for doc in cursor:
@@ -417,60 +418,84 @@ class OplogThread(threading.Thread):
                             last_id = doc["_id"]
                             yield doc
                         break
-                    except pymongo.errors.AutoReconnect:
+                    except (pymongo.errors.AutoReconnect,
+                            pymongo.errors.OperationFailure):
                         attempts += 1
                         time.sleep(1)
 
+        def upsert_each(dm):
+            num_inserted = 0
+            num_failed = 0
+            for num, doc in enumerate(docs_to_dump()):
+                if num % 10000 == 0:
+                    logging.debug("Upserted %d docs." % num)
+                try:
+                    dm.upsert(doc)
+                    num_inserted += 1
+                except Exception:
+                    if self.continue_on_error:
+                        logging.exception(
+                            "Could not upsert document: %r" % doc)
+                        num_failed += 1
+                    else:
+                        raise
+            logging.debug("Upserted %d docs" % num_inserted)
+            if num_failed > 0:
+                logging.error("Failed to upsert %d docs" % num_failed)
+
+        def upsert_all(dm):
+            try:
+                dm.bulk_upsert(docs_to_dump())
+            except Exception as e:
+                if self.continue_on_error:
+                    logging.exception("OplogThread: caught exception"
+                                      " during bulk upsert, re-upserting"
+                                      " documents serially")
+                    upsert_each(dm)
+                else:
+                    raise
+
+        def do_dump(dm, error_queue):
+            try:
+                # Bulk upsert if possible
+                if hasattr(dm, "bulk_upsert"):
+                    logging.debug("OplogThread: Using bulk upsert function for "
+                                  "collection dump")
+                    upsert_all(dm)
+                else:
+                    logging.debug(
+                        "OplogThread: DocManager %s has no "
+                        "bulk_upsert method.  Upserting documents "
+                        "serially for collection dump." % str(dm))
+                    upsert_each(dm)
+            except:
+                # Likely exceptions:
+                # pymongo.errors.OperationFailure,
+                # mongo_connector.errors.ConnectionFailed
+                # mongo_connector.errors.OperationFailed
+                error_queue.put(sys.exc_info())
+
+
         # Extra threads (if any) that assist with collection dumps
         dumping_threads = []
         # Did the dump succeed for all target systems?
         dump_success = True
         # Holds any exceptions we can't recover from
         errors = queue.Queue()
-        try:
+
+        if len(self.doc_managers) == 1:
+            do_dump(self.doc_managers[0], errors)
+        else:
+            # Slight performance gain breaking dump into separate
+            # threads if > 1 replication target
             for dm in self.doc_managers:
-                # Bulk upsert if possible
-                if hasattr(dm, "bulk_upsert"):
-                    logging.debug("OplogThread: Using bulk upsert function for"
-                                  "collection dump")
-                    # Slight performance gain breaking dump into separate
-                    # threads, only if > 1 replication target
-                    if len(self.doc_managers) == 1:
-                        dm.bulk_upsert(docs_to_dump())
-                    else:
-                        def do_dump(error_queue):
-                            all_docs = docs_to_dump()
-                            try:
-                                dm.bulk_upsert(all_docs)
-                            except Exception:
-                                # Likely exceptions:
-                                # pymongo.errors.OperationFailure,
-                                # mongo_connector.errors.ConnectionFailed
-                                # mongo_connector.errors.OperationFailed
-                                error_queue.put(sys.exc_info())
-
-                        t = threading.Thread(target=do_dump, args=(errors,))
-                        dumping_threads.append(t)
-                        t.start()
-                else:
-                    logging.debug("OplogThread: DocManager %s has not"
-                                  "bulk_upsert method.  Upserting documents "
-                                  "serially for collection dump." % str(dm))
-                    num = 0
-                    for num, doc in enumerate(docs_to_dump()):
-                        if num % 10000 == 0:
-                            logging.debug("Upserted %d docs." % num)
-                        dm.upsert(doc)
-                    logging.debug("Upserted %d docs" % num)
-
+                t = threading.Thread(target=do_dump, args=(dm, errors))
+                dumping_threads.append(t)
+                t.start()
             # cleanup
             for t in dumping_threads:
                 t.join()
 
-        except Exception:
-            # See "likely exceptions" comment above
-            errors.put(sys.exc_info())
-
         # Print caught exceptions
         try:
             while True:
@@ -513,26 +538,64 @@ class OplogThread(threading.Thread):
 
         The cursor is set to either the beginning of the oplog, or
         wherever it was last left off.
+
+        Returns the cursor and the number of documents left in the cursor.
         """
-        logging.debug("OplogThread: Initializing the oplog cursor.")
         timestamp = self.read_last_checkpoint()
 
-        if timestamp is None and self.collection_dump:
-            timestamp = self.dump_collection()
-            if timestamp:
-                msg = "Dumped collection into target system"
-                logging.info('OplogThread: %s %s'
-                             % (self.oplog, msg))
-        elif timestamp is None:
-            # set timestamp to top of oplog
-            timestamp = retry_until_ok(self.get_last_oplog_timestamp)
+        if timestamp is None:
+            if self.collection_dump:
+                # dump collection and update checkpoint
+                timestamp = self.dump_collection()
+                if timestamp is None:
+                    return None, 0
+            else:
+                # Collection dump disabled:
+                # return cursor to beginning of oplog.
+                cursor = self.get_oplog_cursor()
+                self.checkpoint = self.get_last_oplog_timestamp()
+                self.update_checkpoint()
+                return cursor, retry_until_ok(cursor.count)
 
         self.checkpoint = timestamp
-        cursor = self.get_oplog_cursor(timestamp)
-        if cursor is not None:
-            self.update_checkpoint()
+        self.update_checkpoint()
 
-        return cursor
+        for i in range(60):
+            cursor = self.get_oplog_cursor(timestamp)
+            cursor_len = retry_until_ok(cursor.count)
+
+            if cursor_len == 0:
+                # rollback, update checkpoint, and retry
+                logging.debug("OplogThread: Initiating rollback from "
+                              "get_oplog_cursor")
+                self.checkpoint = self.rollback()
+                self.update_checkpoint()
+                return self.init_cursor()
+
+            # try to get the first oplog entry
+            try:
+                first_oplog_entry = next(cursor)
+            except StopIteration:
+                # It's possible for the cursor to become invalid
+                # between the cursor.count() call and now
+                time.sleep(1)
+                continue
+
+            # first entry should be last oplog entry processed
+            cursor_ts_long = util.bson_ts_to_long(
+                first_oplog_entry.get("ts"))
+            given_ts_long = util.bson_ts_to_long(timestamp)
+            if cursor_ts_long > given_ts_long:
+                # first entry in oplog is beyond timestamp
+                # we've fallen behind
+                return None, 0
+
+            # first entry has been consumed
+            return cursor, cursor_len - 1
+
+        else:
+            raise errors.MongoConnectorError(
+                "Could not initialize oplog cursor.")
 
     def update_checkpoint(self):
         """Store the current checkpoint in the oplog progress dictionary.
@@ -628,10 +691,16 @@ class OplogThread(threading.Thread):
                 obj_id = bson.objectid.ObjectId
                 bson_obj_id_list = [obj_id(doc['_id']) for doc in doc_list]
 
+                find_kwargs = {
+                    'fields': self._fields,
+                    'timeout': False
+                }
+                if not self.is_sharded:
+                    find_kwargs['exhaust'] = True
                 to_update = util.retry_until_ok(
                     self.main_connection[database][coll].find,
                     {'_id': {'$in': bson_obj_id_list}},
-                    fields=self.fields
+                    **find_kwargs
                 )
                 #doc list are docs in target system, to_update are
                 #docs in mongo
